#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include "define.h"
#include "struct.h"


/*
 *   COMMANDS
 */


class command_ident
{
 public:
  const char*      name;
  do_func*     function;
};


command_ident command_list [] =
{
  { "uptime",      do_uptime     },
  { "settime",     do_settime    },
  { "cangroup",    do_cangroup   },
  { "credits",     do_credits    },
  { "nosell",      do_nosell     },
  { "qflag",       do_qflag      },
  { "typolist",    do_typolist   },
  { "newbiechan",  do_newbiechan },
  { "csocial",     do_csocial    },
  { "clipboard",   do_clipboard  },
  { "finger",      do_finger     },
  { "news",        do_news       },
  { "idletime",    do_idletime   },
  { "avatarchan",  do_avatar     },
  { "ansi",        do_ansi       }, /* TEMPORARY */
  { "settitle",    do_settitle   },
  { "icq",         do_icq        },
  { "/"       ,    do_codechan   },
  { "codechan",    do_codechan   },
  { "afk",         do_afk        },
  { "newareas",    do_newareas   },
  { "wizlist",     do_wizlist    },
  { "bounty",      do_bounty     },
  { "mdelrange",   do_mdelrange  },
  { "plist",       do_plist      },
  { "rdel",        do_rdel       },
  { "at",          do_at         },
  { "rat",         do_rat        },
  { "rnew",        do_rnew       },
  { "-",           do_avatar     },
  { "'",           do_say        },       
  { ":",           do_buildchan  }, 
  { "]",           do_immtalk    },   
  { "}",           do_god        },       
  { "accounts",    do_accounts   },
  { "allegiance",  do_allegiance },
  { "ask",         do_ask        },       
  { "auction",     do_auction    },  
  { "atalk",       do_atalk      }, 
  { "berserk",     do_berserk    },
  { "bid",         do_bid        },
  { "bite",        do_bite       },
  { "build",       do_build      },
  { "bury",        do_bury       },
  { "buy",         do_buy        },       
  { "calculate",   do_calculate  },
  { "cedit",       do_cedit      }, 
  { "changes",     do_changes    },
  { "chant",       do_chant      },     
  { "chat",        do_chat       },      
  { "clone",       do_clone      },    
  { "close",       do_close      },    
  { "configure",   do_configure  },
  { "consent",     do_consent    },
  { "cook",        do_cook       },
  { "credits",     do_credits    },
  { "ctell",       do_ctell      },
  { "custom",      do_custom     },    
  { "ddatabase",   do_ddatabase  },
  { "define",      do_define     },
  { "dig",         do_dig        },
  { "dip",         do_dip        },
  { "disguise",    do_disguise   },
  { "dismount",    do_dismount   },
  { "down",        do_down       },
  { "drink",       do_drink      },    
  { "drop",        do_drop       },     
  { "east",        do_east       },
  { "emote",       do_emote      },     
  { "empty",       do_empty      },    
  { "engage",      do_engage     },
  { "enter",       do_enter      },
  { "equipment",   do_equipment  },
  { "exits",       do_exits      },
  { "exp",         do_exp        },
  { "extract",     do_extract    },
  { "filter",      do_filter     },    
  { "gamestat",    do_gamestat   },
  { "garrote",     do_garrote    },
  { "get",         do_get        },      
  { "give",        do_give       },  
  { "gossip",      do_gossip     },    
  { "gouge",       do_gouge      },
  { "grab",        do_wear       },     
  { "gtell",       do_gtell      },     
  { "hbug",        do_hbug       },    
  { "homepage",    do_homepage   },
  { "identity",    do_identity   },  
  { "iflag",       do_iflag      },     
  { "ignite",      do_ignite     },
  { "index",       do_index      },     
  { "info",        do_info       },      
  { "introduce",   do_introduce  }, 
  { "kick",        do_kick       },       
  { "kill",        do_kill       },       
  { "clans",       do_clans      },
  { "knock",       do_knock      },
  { "label",       do_label      },
  { "language",    do_language   },  
  { "ledit",       do_ledit      },
  { "lie",         do_lie        },
  { "list",        do_list       },      
  { "load",        do_load       },
  { "lset",        do_lset       },      
  { "lstat",       do_lstat      },
  { "marmor",      do_marmor     },
  { "mdesc",       do_mdesc      }, 
  { "medit",       do_medit      },
  { "meditate",    do_meditate   },  
  { "melt",        do_melt       },      
  { "melee",       do_melee      },
  { "message",     do_message    },
  { "mfind",       do_mfind      },
  { "mflag",       do_mflag      },
  { "mload",       do_mload      },
  { "mlog",        do_mlog       }, 
  { "mount",       do_mount      },     
  { "mpflag",      do_mpflag     },
  { "mreset",      do_mreset     },
  { "mset",        do_mset       },
  { "msetrange",   do_msetrange  },
  { "mstat",       do_mstat      }, 
  { "mwhere",      do_mwhere     },  
  { "north",       do_north      },
  { "nostolen",    do_nostolen   },     
  { "ooc",         do_ooc        },
  { "pbug",        do_pbug       },
  { "petlist",     do_petlist    },
  { "purchase",    do_purchase   }, 
  { "pray",        do_pray       },      
  { "prompt",      do_prompt     },
  { "ps",          do_ps         },
  { "qstat",       do_qstat      },
  { "repair",      do_repair     },    
  { "reply",       do_reply      },     
  { "reset",       do_reset      },    
  { "rest",        do_rest       },      
  { "reboot",      do_reboot     },
  { "rename",      do_rename     },    
  { "request",     do_request    },
  { "review",      do_review     },
  { "rmwhere",     do_rmwhere    },   
  { "roomlist",    do_roomlist   },
  { "rtable",      do_rtable     },
  { "rtwhere",     do_rtwhere    },
  { "say",         do_say        },       
  { "sell",        do_sell       },      
  { "shout",       do_shout      },     
  { "sit",         do_sit        },       
  { "sleep",       do_sleep      },     
  { "smoke",       do_smoke      },
  { "smother",     do_smother    },
  { "south",       do_south      },     
  { "split",       do_split      },
  { "stand",       do_stand      },     
  { "tell",        do_tell       },      
  { "to",          do_to         },
  { "typo",        do_typo       },
  { "up",          do_up         },        
  { "value",       do_value      },     
  { "vote",        do_vote       },
  { "wake",        do_wake       },      
  { "wear",        do_wear       },     
  { "west",        do_west       },      
  { "whisper",     do_whisper    },
  { "wield",       do_wear       },     
  { "yell",        do_yell       },      

  { "relations",   do_relations  }, 
  { "eat",         do_eat        },      
  { "fill",        do_fill       },     
  { "hold",        do_wear       },     
  { "inventory",   do_inventory  },
  { "junk",        do_junk       },     
  { "lock",        do_lock       },     
  { "open",        do_open       },     
  { "pick",        do_pick       },     
  { "put",         do_put        },      
  { "quaff",       do_quaff      },    
  { "remove",      do_remove     },   
  { "sacrifice",   do_sacrifice  },
  { "unlock",      do_unlock     },   
  { "assist",      do_assist     },     
  { "backstab",    do_backstab   },   
  { "bash",        do_bash       },      
  { "charge",      do_charge     }, 
  { "consider",    do_consider   },   
  { "disarm",      do_disarm     },     
  { "flee",        do_flee       },       
  { "glance",      do_glance     },     
  { "punch",       do_punch      },      
  { "rescue",      do_rescue     },     
  { "shoot",       do_shoot      },      
  { "spin",        do_spin_kick  },  
  { "throw",       do_throw      },      
  { "wimpy",       do_wimpy      },      
  { "cast",        do_cast       },      
  { "focus",       do_focus      },     
  { "leech",       do_leech      },     
  { "polymorph",   do_polymorph  }, 
  { "prepare",     do_prepare    },   
  { "probe",       do_probe      },     
  { "recite",      do_recite     },    
  { "spells",      do_spells     },    
  { "sweep",       do_sweep      },     
  { "switch",      do_switch     },    
  { "return",      do_return     },    
  { "zap",         do_zap        },       
  { "alias",       do_alias      },     
  { "appearance",  do_appearance },
  { "areas",       do_areas      },     
  { "befriend",    do_befriend   },  
  { "commands",    do_commands   },  
  { "compare",     do_compare    },   
  { "color",       do_color      },     
  { "description", do_descript   },  
  { "help",        do_help       },      
  { "keywords",    do_keywords   },  
  { "last",        do_last       },      
  { "look",        do_look       },      
  { "motd",        do_motd       },      
  { "options",     do_options    },   
  { "peek",        do_peek       },      
  { "pets",        do_pets       },      
  { "qlook",       do_qlook      },     
  { "quests",      do_quests     },    
  { "qwho",        do_qwho       },      
  { "reputation",  do_reputation },
  { "score",       do_score      },     
  { "socials",     do_socials    },   
  { "statistics",  do_statistics },
  { "time",        do_time       },      
  { "who",         do_who        },
  { "rwho",        do_rwho       },       
  { "whois",       do_whois      },     
  { "mail",        do_mail       },      
  { "notes",       do_notes      },     
  { "title",       do_title      },     
  { "follow",      do_follow     },  
  { "group",       do_group      },
  { "name",        do_name       }, 
  { "order",       do_order      },
  { "delete",      do_delete     },    
  { "password",    do_password   },  
  { "quit",        do_quit       },      
  { "save",        do_save       },      
  { "balance",     do_balance    },   
  { "deposit",     do_deposit    },   
  { "withdraw",    do_withdraw   },  
  { "abilities",   do_abilities  }, 
  { "appraise",    do_appraise   },  
  { "bandage",     do_bandage    },   
  { "climb",       do_climb      },     
  { "hands",       do_hands      },      
  { "hide",        do_hide       },       
  { "inspect",     do_inspect    },    
  { "practice",    do_practice   },   
  { "scan",        do_scan       },       
  { "search",      do_search     },     
  { "skin",        do_skin       },       
  { "skills",      do_skills     },     
  { "sneak",       do_sneak      },      
  { "steal",       do_steal      },      
  { "heist",       do_heist      },
  { "energize",    do_energize   },
  { "track",       do_track      },      
  { "untrap",      do_untrap     },     
  { "advance",     do_advance    },   
  { "approve",     do_approve    },   
  { "ban",         do_ban        },       
  { "beep",        do_beep       },      
  { "bamfin",      do_bamfin     },    
  { "bamfout",     do_bamfout    },   
  { "bugs",        do_bugs       },      
  { "disconnect",  do_disconnect },
  { "identify",    do_identify   },  
  { "pardon",      do_pardon     },    
  { "reimburse",   do_reimburse  }, 
  { "where",       do_where      },     
  { "tedit",       do_tedit      },     
  { "tflag",       do_tflag      },     
  { "tset",        do_tset       },     
  { "tstat",       do_tstat      },       
  { "dedit",       do_dedit      },       
  { "dflag",       do_dflag      },       
  { "dset",        do_dset       },        
  { "dstat",       do_dstat      },       
  { "rbug",        do_rbug       },        
  { "redit",       do_redit      },       
  { "rdesc",       do_rdesc      },       
  { "rflag",       do_rflag      },       
  { "rlog",        do_rlog       },        
  { "rname",       do_rname      },       
  { "rset",        do_rset       },        
  { "rstat",       do_rstat      },       
  { "odesc",       do_odesc      },   
  { "oedit",       do_oedit      },   
  { "oextra",      do_oextra     },  
  { "ofind",       do_ofind      },   
  { "oflag",       do_oflag      },   
  { "olist",       do_olist      },   
  { "oload",       do_oload      },   
  { "olog",        do_olog       },    
  { "oset",        do_oset       },    
  { "ostat",       do_ostat      },   
  { "owhere",      do_owhere     },  
  { "opedit",      do_opedit     },  
  { "opflag",      do_opflag     },  
  { "opcode",      do_opcode     },  
  { "opdata",      do_opdata     },  
  { "opstat",      do_opstat     },  
  { "opset",       do_opset      },   
  { "hdesc",       do_hdesc      },   
  { "hedit",       do_hedit      },   
  { "hset",        do_hset       },    
  { "acode",       do_acode      },   
  { "adata",       do_adata      },   
  { "aedit",       do_aedit      },   
  { "aflag",       do_aflag      },   
  { "aset",        do_aset       },    
  { "astat",       do_astat      },   
  { "mpedit",      do_mpedit     },  
  { "mpcode",      do_mpcode     },  
  { "mpdata",      do_mpdata     },  
  { "mpstat",      do_mpstat     },  
  { "mpset",       do_mpset      },   
  { "cflag",       do_cflag      },   
  { "qedit",       do_qedit      },   
  { "qremove",     do_qremove    }, 
  { "qset",        do_qset       },    
  { "qwhere",      do_qwhere     },  
  { "shcustom",    do_shcustom   },
  { "shedit",      do_shedit     },  
  { "shflag",      do_shflag     },    
  { "affects",     do_affects    },   
  { "high",        do_high       },      
  { "level",       do_level      },     
  { "memory",      do_memory     },    
  { "move",        do_move       },      
  { "pull",        do_pull       },      
  { "push",        do_push       },      
  { "read",        do_read       },      
  { "system",      do_system     },    
  { "wanted",      do_wanted     },    
  { "whistle",     do_whistle    },   
  { "functions",   do_functions  },
  { "return",      do_return     },   
  { "deny",        do_deny       },     
  { "echo",        do_echo       },     
  { "force",       do_force      },    
  { "freeze",      do_freeze     },   
  { "god",         do_god        },      
  { "goto",        do_goto       },     
  { "holylight",   do_holylight  },
  { "immtalk",     do_immtalk    },  
  { "imprison",    do_imprison   }, 
  { "invis",       do_invis      },    
  { "lag",         do_lag        },      
  { "map",         do_map        },      
  { "peace",       do_peace      },    
  { "purge",       do_purge      },    
  { "recho",       do_recho      },    
  { "restore",     do_restore    },  
  { "rowhere",     do_rowhere    },  
  { "shutdown",    do_shutdown   }, 
  { "snoop",       do_snoop      },    
  { "slay",        do_slay       },     
  { "sset",        do_sset       },     
  { "tfind",       do_tfind      },    
  { "train",       do_train      },    
  { "transfer",    do_transfer   }, 
  { "trust",       do_trust      },    
  { "users",       do_users      },    
  { "vaporise",    do_vaporise   },
  { "wizlock",     do_wizlock    },  
  { "write",       do_write      },    
  { "",            NULL          }
};


void init_commands( )
{
  int i, j;

  for( i = 0; i < MAX_ENTRY_COMMAND; i++ ) {
    for( j = 0; command_list[j].name[0] != '\0'; j++ ) {
      if( !strcasecmp( command_table[i].func_name, command_list[j].name ) ) {
        command_table[i].function = command_list[j].function;
        break;
        }
      }
    }

  return;
}


/*
 *   SPELLS 
 */


class spell_ident
{
 public:
  const char*       name;
  spell_func*   function;   
  bool              cast;  /* Can the spell be cast?  Most likely yes */
  bool         proc_cast;  /* Can the spell be cast by online procs? */
  bool            recite;  /* Used in scrolls? */
  bool               zap;  /* Used in wands? */
  bool             quaff;  /* Used in potions? */
  bool      throw_potion;  /* Thrown as a potion? */
  bool             drink;  /* Have an effect when drunk? */
  bool      throw_liquid;  /* Have and effect when thrown as liquid? */
  bool               dip;  /* Have an effect when dipped things into? */
  bool              fill;  /* Have an effect when filled from? */
};


spell_ident spell_list [] =
{                                               /*  N  C  R  Z  Q  Tp D  Tl Dp F  */
  { "acid blast",         spell_acid_blast,         1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
  { "acid storm",         spell_acid_storm,         1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "armor",              spell_armor,              1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "astral gate",        spell_astral_gate,        1, 0, 1, 0, 0, 0, 0, 0, 0, 0 },
  { "augury",             spell_augury,             1, 1, 1, 1, 0, 0, 0, 0, 1, 0 },
  { "balm",               spell_balm,               1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "banishment",         spell_banishment,         1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "barkskin",           spell_barkskin,           1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "bless",              spell_bless,              1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "blind",              spell_blind,              1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "blink",              spell_blink,              1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "burning hands",      spell_burning_hands,      1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "calm",               spell_calm,               1, 1, 1, 1, 0, 1, 0, 1, 0, 0 },
  { "call lightning",     spell_call_lightning,     1, 1, 1, 1, 0, 1, 0, 1, 0, 0 },
  { "cause critical",     spell_cause_critical,     1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "cause light",        spell_cause_light,        1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "cause serious",      spell_cause_serious,      1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "conflagration",      spell_conflagration,      1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "confuse",            spell_confuse,            1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "construct golem",    spell_construct_golem,    1, 0, 1, 1, 0, 0, 0, 0, 1, 0 },
  { "create feast",       spell_create_feast,       1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "create food",        spell_create_food,        1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "create light",       spell_create_light,       1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "create water",       spell_create_water,       1, 0, 1, 1, 0, 0, 1, 1, 1, 1 },
  { "cure blindness",     spell_cure_blindness,     1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "cure critical",      spell_cure_critical,      1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "cure light",         spell_cure_light,         1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "cure poison",        spell_cure_poison,        1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "cure serious",       spell_cure_serious,       1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "detect invisible",   spell_detect_invisible,   1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "displace",           spell_displace,           1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "eagle eye",          spell_eagle_eye,          1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "fear",               spell_fear,               1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "find familiar",      spell_find_familiar,      1, 0, 1, 1, 0, 0, 0, 0, 1, 0 },
  { "fire shield",        spell_fire_shield,        1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "fireball",           spell_fireball,           1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "flame strike",       spell_flame_strike,       1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "float",              spell_float,              1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "fly",                spell_fly,                1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "greater animation",  spell_greater_animation,  1, 0, 1, 1, 0, 0, 0, 0, 1, 0 },
  { "hallucinate",        spell_hallucinate,        1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "harm",               spell_harm,               1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "haste",              spell_haste,              1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "heal",               spell_heal,               1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "identify",           spell_identify,           1, 0, 1, 1, 0, 0, 0, 0, 1, 0 },
  { "invulnerability",    spell_invulnerability,    1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "lesser summoning",   spell_lesser_summoning,   1, 0, 1, 1, 0, 0, 0, 0, 1, 0 },
  { "lightning bolt",     spell_lightning_bolt,     1, 1, 1, 1, 0, 1, 0, 1, 0, 0 },
  { "magic missile",      spell_magic_missile,      1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "maelstrom",          spell_maelstrom,          1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "major enchantment",  spell_major_enchantment,  1, 0, 1, 1, 0, 0, 0, 0, 1, 0 },
  { "mind blade",         spell_mind_blade,         1, 1, 1, 1, 0, 1, 0, 1, 0, 0 },
  { "minor enchantment",  spell_minor_enchantment,  1, 0, 1, 1, 0, 0, 0, 0, 1, 0 },
  { "neutralize",         spell_neutralize,         1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "pass door",          spell_pass_door,          1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "polymorph",          spell_polymorph,          1, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
  { "poultice",           spell_poultice,           1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "prismic missile",    spell_prismic_missile,    1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "protection/good",    spell_protect_good,       1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "protection/evil",    spell_protect_evil,       1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "recall",             spell_recall,             1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "remove curse",       spell_remove_curse,       1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "request ally",       spell_request_ally,       1, 0, 1, 1, 0, 0, 0, 0, 1, 0 },
  { "resist acid",        spell_resist_acid,        1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "resist cold",        spell_resist_cold,        1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "resist fire",        spell_resist_fire,        1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "resist shock",       spell_resist_shock,       1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "resurrect",          spell_resurrect,          1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "revitalize",         spell_revitalize,         1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "sanctify",           spell_sanctify,           1, 0, 1, 1, 0, 0, 0, 0, 1, 1 },
  { "scry",               spell_scry,               1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "sense danger",       spell_sense_danger,       1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "slay",               spell_slay,               1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "shock",              spell_shock,              1, 1, 1, 1, 0, 1, 0, 1, 0, 0 },
  { "sleep",              spell_sleep,              1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "slow",               spell_slow,               1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "summon",             spell_summon,             1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "surcease",           spell_surcease,           1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "transfer",           spell_transfer,           1, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, 
  { "restoration",        spell_restoration,        1, 1, 1, 1, 1, 1, 1, 1, 0, 0 }, 
  { "turn undead",        spell_turn_undead,        1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "vitality",           spell_vitality,           1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "water breathing",    spell_water_breathing,    1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "web",                spell_web,                1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "wizard lock",        spell_wizard_lock,        1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "youth",              spell_youth,              1, 1, 1, 1, 1, 1, 1, 1, 1, 0 },
                                                /*  N  C  R  Z  Q  Tp D  Tl Dp F  */
  { "holy wrath",         spell_holy_wrath,         1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "tame",               spell_tame,               1, 1, 1, 1, 0, 1, 0, 1, 0, 0 },
  { "faerie fire",        spell_faerie_fire,        1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "invisibility",       spell_invisibility,       1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "protection plants",  spell_protection_plants,  1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "curse",              spell_curse,              1, 1, 1, 1, 1, 1, 1, 1, 0, 0 }, 
  { "mists of sleep",     spell_mists_sleep,        1, 1, 1, 1, 0, 1, 0, 1, 0, 0 },
  { "locust swarm",       spell_locust_swarm,       1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "poison cloud",       spell_poison_cloud,       1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "magic mapping",      spell_magic_mapping,      1, 1, 1, 0, 0, 0, 0, 0, 0, 0 },
  { "amnesia",            spell_amnesia,            1, 1, 1, 0, 1, 0, 1, 0, 0, 0 },
  { "detect hidden",      spell_detect_hidden,      1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "detect good",        spell_detect_good,        1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "detect evil",        spell_detect_evil,        1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "ignite weapon",      spell_ignite_weapon,      1, 0, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "chain lightning",    spell_chain_lightning,    1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "infravision",        spell_infravision,        1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "mystic shield",      spell_mystic_shield,      1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "animate dead",       spell_animate_dead,       1, 0, 1, 1, 0, 0, 0, 0, 1, 0 },
  { "ogre strength",      spell_ogre_strength,      1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "silence",            spell_silence,            1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "drain life",         spell_drain_life,         1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "blinding light",     spell_blinding_light,     1, 1, 1, 1, 0, 0, 0, 0, 1, 0 },
  { "gift of tongues",    spell_gift_of_tongues,    1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "protect life",       spell_protect_life,       1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "true sight",         spell_true_sight,         1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "sense life",         spell_sense_life,         1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "detect magic",       spell_detect_magic,       1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "continual light",    spell_continual_light,    1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "cure disease",       spell_cure_disease,       1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "animate clay",       spell_animate_clay,       1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
  { "freeze",             spell_freeze,             1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "chilling touch",     spell_chilling_touch,     1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "ice storm",          spell_ice_storm,          1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "replicate",          spell_replicate,          1, 0, 1, 1, 0, 0, 0, 0, 1, 0 },
  { "thorn shield",       spell_thorn_shield,       1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "group serious",      spell_group_serious,      1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "group critical",     spell_group_critical,     1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "ion shield",         spell_ion_shield,         1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "meteor swarm",       spell_meteor_swarm,       1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "ice lance",          spell_ice_lance,          1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "purify",             spell_purify,             1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "wither",             spell_wither,             1, 1, 1, 1, 1, 1, 1, 1, 0, 0 },
  { "obscure",            spell_obscure,            1, 0, 1, 1, 0, 0, 0, 0, 1, 0 },
  { "conjure elemental",  spell_conjure_elemental,  1, 0, 1, 1, 0, 0, 0, 0, 1, 0 },
  { "find mount",         spell_find_mount,         1, 0, 1, 1, 0, 0, 0, 0, 1, 0 },
  { "hawks view",         spell_hawks_view,         1, 1, 1, 1, 0, 0, 0, 0, 0, 0 },
  { "",                   NULL,                     1, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
};                                              /*  N  C  R  Z  Q  Tp Dk Tl Dp F  */


/* PUIOK 24th July 2000 */
void init_usable_flags( spell_data* data, spell_ident* ident )
{
  int flags[10] = { ident->cast, ident->proc_cast, ident->recite, ident->zap,
                    ident->quaff, ident->throw_potion, ident->drink,
                    ident->throw_liquid, ident->dip, ident->fill };

  data->usable_flag = 0;
  for( int i = 0; i < 10; i++ )
    if( flags[i] == 1 )
      set_bit( &data->usable_flag, i );
}


void init_spells( )
{
  int i, j;

  for( i = 0; i < MAX_SPELL; i++ ) {
    spell_table[i].function = NULL; 
    for( j = 0; spell_list[j].name[0] != '\0'; j++ ) {
      if( !strcasecmp( spell_table[i].name, spell_list[j].name ) ) {
        spell_table[i].function = spell_list[j].function;
        init_usable_flags( &spell_table[i], &spell_list[j] );
        break;
        }
      }
    if( spell_table[i].function == NULL ) 
      panic( "Init_Spells: Null function - %s", spell_table[i].name );
    }

  return;
}


